import googlemaps
import math

class CentroidCalculator:
    def __init__(self, api_key):
        """
        Initializes the CentroidCalculator object with the Google Maps API key.
        """
        self.gmaps = googlemaps.Client(key=api_key)

    def geocode_address(self, address):
        """
        Geocodes an address to get the latitude and longitude coordinates.
        @parameters:
        - address (str): The address to geocode.
        @returns:
        - [lat, lng] (list): The latitude and longitude coordinates of the address.
        """
        result = self.gmaps.geocode(address)
        if result and 'geometry' in result[0]:
            location = result[0]['geometry']['location']
            lat = location['lat']
            lng = location['lng']
            return [lat, lng]
        else:
            print("Could not geocode address (Skipping): ", address)
            return None

    def calculate_two_point_centroid(self, lat1, lng1, w1, lat2, lng2, w2):
        """
        Calculates the centroid of two points with weights.
        @parameters:
        - lat1 (float): The latitude of the first point.
        - lng1 (float): The longitude of the first point.
        - w1 (int): The weight of the first point.
        - lat2 (float): The latitude of the second point.
        - lng2 (float): The longitude of the second point.
        - w2 (int): The weight of the second point.
        @returns:
        - [clat, clng, w1 + w2] (list): The latitude, longitude, and combined weight of the centroid.
        """
        lat1_rad = lat1 * (math.pi / 180)
        lat2_rad = lat2 * (math.pi / 180)
        lng1_rad = lng1 * (math.pi / 180)
        lng2_rad = lng2 * (math.pi / 180)
        delta_lat_rad = (lat2 - lat1) * (math.pi / 180)
        delta_lng_rad = (lng2 - lng1) * (math.pi / 180)
        h = (pow(math.sin(delta_lat_rad / 2), 2) + (math.cos(lat1_rad) * math.cos(lat2_rad) * pow(math.sin(delta_lng_rad / 2), 2)))
        c = 2 * math.atan2(math.sqrt(h), math.sqrt(1 - h))
        if w1 < w2:
            f = 1 - w1 / (w1 + w2)
        else:
            f = 1 - w2 / (w1 + w2)
        a = math.sin((1 - f) * c) / math.sin(c)
        b = math.sin(f * c) / math.sin(c)
        x = a * math.cos(lat1_rad) * math.cos(lng1_rad) + b * math.cos(lat2_rad) * math.cos(lng2_rad)
        y = a * math.cos(lat1_rad) * math.sin(lng1_rad) + b * math.cos(lat2_rad) * math.sin(lng2_rad)
        z = a * math.sin(lat1_rad) + b * math.sin(lat2_rad)
        clat_rad = math.atan2(z, math.sqrt(x * x + y * y))
        clng_rad = math.atan2(y, x)
        clat = clat_rad * (180 / math.pi)
        clng = clng_rad * (180 / math.pi)
        return [clat, clng, w1 + w2]

    def generate_centroid(self, geocoded):
        """
        Generates the centroid of a list of geocoded addresses with weights.
        @parameters:
        - geocoded (list): A list of geocoded addresses.
        @returns:
        - [clat, clng, w] (list): The latitude, longitude, and combined weight of the centroid.
        """
        while len(geocoded) > 1:
            new_coord = self.calculate_two_point_centroid(*geocoded[0], *geocoded[1])
            geocoded = geocoded[2:]
            geocoded.append(new_coord)
        return geocoded[0]

    def get_centroid(self, addresses):
        """
        Calculates the centroid of a list of addresses.
        @parameters:
        - addresses (list): A list of addresses.
        Each element in the addresses list can be:
              A string representing the address,
              A geocoded address [lat, lng],
              An address with a weight [address, weight] | [lat, lng, weight].
        @returns:
        - [clat, clng, w] (list): The latitude, longitude, and combined weight of the centroid.
        """
        geocoded =[]

        for address in addresses:
            if type(address) == list:
                if len(address) == 2:
                    if type(address[0]) == str and type(address[1]) == int:
                        geocoded_address = self.geocode_address(address[0])
                        if geocoded_address is None:
                            continue
                        geocoded_address.append(address[1])
                    elif type(address[0]) == float and type(address[1]) == float \
                        and address[0] >= -90.0 and address[0] <= 90.0 and address[1] >= -180.0 and address[1] <= 180.0:
                        geocoded_address = address
                        geocoded_address.append(1)
                    else:
                        print("Could not geocode address (Skipping): ", address)
                        continue
                elif len(address) == 3:
                    if type(address[0]) == float and type(address[1]) == float and type(address[2]) == int \
                        and address[0] >= -90.0 and address[0] <= 90.0 and address[1] >= -180.0 and address[1] <= 180.0:
                        geocoded_address = address
                    else:
                        print("Could not geocode address (Skipping): ", address)
                        continue
                else:
                    print("Could not geocode address (Skipping): ", address)
                    continue
            elif type(address) == str:
                geocoded_address = self.geocode_address(address)
                if geocoded_address is None:
                    continue
                geocoded_address.append(1)
            else:
                print("Could not geocode address (Skipping): ", address)
                continue
            geocoded.append(geocoded_address)

        if len(geocoded) > 1:
            print(geocoded)
            return self.generate_centroid(geocoded)
        else:
            print("Failed to calculate centroid (Not enough geocoded addresses)")
            return None



import unittest
import os
from dotenv import load_dotenv
from app.geocentroid.src.centroid_calculator import CentroidCalculator

load_dotenv()

class TestCentroidCalculator(unittest.TestCase):
    def setUp(self):
        self.calculator = CentroidCalculator(os.environ.get("API_KEY"))
    
    def assertListAlmostEqual(self, list1, list2, tol):
        """
        Asserts that two lists are equal up to a certain tolerance.
        """
        self.assertEqual(len(list1), len(list2))
        for a, b in zip(list1, list2):
            self.assertAlmostEqual(a, b, tol)
    
    # The error tolerance for the coordinates is up to 3 decimal places. 
    # Because at 4 decimal places, the difference is only 11.1m
    def test_geocode_address(self):
        result = self.calculator.geocode_address("64 Marshall st, Waterloo Ontario")
        self.assertListAlmostEqual(result, [43.4741752, -80.5204558], 3)
        
    def test_get_centroid_all_addresses(self):
        address_list_1 = ["64 Marshall st Waterloo Ontario", "CIF waterloo ontario", "King Street towers, waterloo ontario"]
        result_1 = self.calculator.get_centroid(address_list_1)
        self.assertListAlmostEqual(result_1, [43.47649843865154, -80.53145133903517, 3],3)
        
    def test_get_centroid_all_addresses_with_weight(self):
        address_list_2 = [["64 Marshall st Waterloo Ontario", 5], "CIF waterloo ontario", "King Street towers, waterloo ontario"]
        result_2 = self.calculator.get_centroid(address_list_2)
        self.assertListAlmostEqual(result_2, [43.47580075938039, -80.540801226302, 7],3)
        
    def test_get_centroid_geocoded_addresses_with_weight(self):
        address_list_3 = [[43.4740533, -80.5205138, 5], [43.475277, -80.54781349999999, 1], [43.4801632, -80.5260265, 1]]
        result_3 = self.calculator.get_centroid(address_list_3)
        self.assertListAlmostEqual(result_3, [43.47580075938039, -80.540801226302, 7],3)
        
    def test_get_centroid_single_address(self):
        address_list_4 = [[43.4740533, -80.5205138, 5]]
        result_4 = self.calculator.get_centroid(address_list_4)
        self.assertIsNone(result_4)
        
    def test_get_centroid_mix_address_and_geocoded(self):
        address_list_5 = [["64 Marshall st Waterloo Ontario", 5], [43.475277, -80.54781349999999, 1], "King Street towers, waterloo ontario"]
        result_5 = self.calculator.get_centroid(address_list_5)
        self.assertListAlmostEqual(result_5, [43.47580075938039, -80.540801226302, 7],3)
        
    def test_get_centroid_mix_weighted_and_non_weighted(self):
        address_list_6 = ["64 Marshall st Waterloo Ontario", [43.475277, -80.54781349999999, 2], "King Street towers, waterloo ontario"]
        result_6 = self.calculator.get_centroid(address_list_6)
        self.assertListAlmostEqual(result_6, [43.47619329770745, -80.53554194128088, 4],3)
        
    def test_get_centroid_fake_address(self):
        address_list_7 = ["fake address", [43.475277, -80.54781349999999, 2], "King Street towers, waterloo ontario"]
        result_7 = self.calculator.get_centroid(address_list_7)
        self.assertListAlmostEqual(result_7, [43.4785349262818, -80.53328922483436, 3],3)
        
    def test_get_centroid_invalid_latlng_list(self):
        address_list_8 = ["64 Marshall st Waterloo Ontario", [-80.54781349999999], "King Street towers, waterloo ontario"]
        result_8 = self.calculator.get_centroid(address_list_8)
        self.assertListAlmostEqual(result_8, [43.47710828310337, -80.52327001064637, 2],3)
        
    def test_get_centroid_invalid_latlng_coord(self):
        address_list_9 = ["64 Marshall st Waterloo Ontario", [443.475277, -80.54781349999999, 2], "King Street towers, waterloo ontario"]
        result_9 = self.calculator.get_centroid(address_list_9)
        self.assertListAlmostEqual(result_9, [43.47710828310337, -80.52327001064637, 2],3)
        
    def test_get_centroid_invalid_weight(self):
        address_list_10 = [["64 Marshall st Waterloo Ontario", 5.1], [43.475277, -80.54781349999999, 1]]
        result_10 = self.calculator.get_centroid(address_list_10)
        self.assertIsNone(result_10)

